/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* time_t */
#include <sys/stat.h>	/* stat() */
#include <ctype.h>		/* toupper() */
#include <errno.h>		/* errno */
#include <stdio.h>		/* puts() */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcmp() */
#include <time.h>		/* time() */
#include <unistd.h>		/* stat() */

/* readdir() */
#ifdef HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include "clex.h"
#include "completion.h"

#include "cfg.h"		/* config_num() */
#include "control.h"	/* control_loop() */
#include "edit.h"		/* edit_update() */
#include "filepanel.h"	/* files_find() */
#include "inout.h"		/* win_remark() */
#include "panel.h"		/* pan_adjust() */
#include "start.h"		/* startup_warning() */
#include "userdata.h"	/* username_find() */
#include "ustring.h"	/* us_init() */
#include "util.h"		/* emalloc() */

extern int errno;
extern char **environ;

/*
 * commands are stored in several lists depending on
 * the first character. It is simpler then sorting.
 */
#define LISTS	27
#define LIST_NR(ch) ((ch) >= 'a' && (ch) <= 'z' ? (ch) - 'a' : 26)
typedef struct cmd {
	const char *command;	/* command name */
	struct cmd *next;		/* linked list ptr */
} CMD;

/* linked list of PATHDIRs holds info about commands in user's $PATH */
typedef struct pathdir {
	struct pathdir *next;	/* linked list ptr */
	const char *dir;		/* PATH directory name */
	time_t timestamp;		/* time of last directory scan */
	CMD *commands[LISTS];	/* lists of commands in this directory */
} PATHDIR;

typedef struct {
	/* input: completion request data */
	const char *str;	/* string to be completed, might be not null
						   terminated, see 'strlen' */
	size_t strlen;		/* length of 'str' */
	const char *dir;	/* directory to attempt the completion in
						 * (if required by the type of completion) */
	size_t dirlen;		/* length of 'dir' (which is null terminated) */
	short int type;		/* type of completion - one of COMPL_TYPE_XX */

	/* output: completion results */ 
	int cnt;			/* number of completion candidates */
	int err;			/* errno value (if cnt is zero) */
	int cwd;			/* true if 'dir' is the current directory */
	size_t clen;		/* how many characters following 'str' */
						/* are the same for all candidates */
	char first_char[256];	/* map of first characters */
	CANDIDATE *candidate;	/* list of completion candidates */
} COMPLETION;

static PATHDIR *pathdir_list_head = 0;
static COMPLETION compl;
static int cfg_compl;		/* max number of candidates to present */
static size_t *env_size;	/* length of environment variables */

static void
register_command_path(const char *path)
{
	int i;
	PATHDIR *ppd;

	if (*path == '\0')
		path = ".";
	else if (*path != '/' && strcmp(path,".")) {
		puts(
		  "COMPLETION: Command completion does not support relative\n"
		  "            directories in the command search path $PATH.");
		  /* ... and it is a stupid idea anyway */
		startup_warning();
		return;
	}
	ppd = emalloc(sizeof(PATHDIR));
	ppd->dir = path;
	ppd->timestamp = 0;
	for (i = 0; i < LISTS; i++)
		ppd->commands[i] = 0;
	ppd->next = pathdir_list_head;
	pathdir_list_head = ppd;
}

static void
command_completion_init(void)
{
	char *path, *dir;

	if ( (path = getenv("PATH")) == 0)
		return;

	/* split PATH to components */
	for (dir = path = estrdup(path); *path; path++)
		if (*path == ':') {
			*path = '\0';
			register_command_path(dir);
			dir = path + 1;
		}
	register_command_path(dir);
}

static void
command_completion_refresh(void)
{
	int list;
	time_t now;
	const char *file;
	struct stat st;
	struct dirent *direntry;
	DIR *dd;
	CMD *pc;
	PATHDIR *ppd;

	now = time(0);
	for (ppd = pathdir_list_head; ppd; ppd = ppd->next) {
		if (*ppd->dir != '/')
			continue;
		if (stat(ppd->dir,&st) == 0 && st.st_mtime < ppd->timestamp)
			continue;
		for (list = 0; list < LISTS; list++)
			while ( (pc = ppd->commands[list]) ) {
				ppd->commands[list] = pc->next;
				free(pc);
			}

		ppd->timestamp = now;
		if ( (dd = opendir(ppd->dir)) == 0) {
			ppd->timestamp = 0;
			continue;
		}
		pathname_set_directory(ppd->dir);
		while ( (direntry = readdir(dd)) ) {
			file = direntry->d_name;
			if (stat(pathname_join(file),&st) < 0 ||
			  !S_ISREG(st.st_mode) ||
			  (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) == 0)
				continue;
			list = LIST_NR(*file);
			pc = emalloc(sizeof(CMD));
			pc->command = estrdup(file);
			pc->next = ppd->commands[list];
			ppd->commands[list] = pc;
		}
		closedir(dd);
	}
}

static void
environ_completion_init(void)
{
	int i;
	char *pch;

	for (i = 0; environ[i]; i++)
		/* just counting */;
	if (i == 0)
		return;	/* no environment ?! */

	env_size = emalloc(i * sizeof(size_t));
	while (--i >= 0) {
		pch = strchr(environ[i],'=');	/* VARIABLE=VALUE */
		env_size[i] = pch ? pch - environ[i] : 0;
	}
}

void
completion_initialize(void)
{
	int i;

	command_completion_init();
	command_completion_refresh();
	environ_completion_init();

	cfg_compl = config_num(CFG_COMPLETIONS);
	panel_compl.type = PANEL_TYPE_COMPL;
	panel_compl.min = -1;
	panel_compl.candidate = compl.candidate =
	  emalloc(cfg_compl * sizeof(CANDIDATE));
	for (i = 0; i < cfg_compl; i++)
		US_INIT(compl.candidate[i].name);
}

static int
qcmp_cand(const void *e1, const void *e2)
{
	return strcmp(
	  USTR(((CANDIDATE *)e1)->name),
	  USTR(((CANDIDATE *)e2)->name)
	);
}

void
compl_prepare(void)
{
	panel_compl.cnt = compl.cnt;
	panel_compl.top = panel_compl.curs = -1;
	panel_compl.compl_type = compl.type;

	/* note: user names are already sorted */
	if (panel_compl.compl_type != COMPL_TYPE_USER)
		qsort(compl.candidate,compl.cnt,sizeof(CANDIDATE),qcmp_cand);

	panel = (PANEL *)&panel_compl;
}

static void
register_candidate(const char *cand, int isdir)
{
	int i;
	size_t c;
	static const char *cand0;

	if (compl.cnt < cfg_compl) {
		if (compl.type == COMPL_TYPE_PATHCMD)
			/* check for duplicates, like awk in /bin and /usr/bin */
			for (i = 0; i < compl.cnt; i++)
				if (strcmp(USTR(compl.candidate[i].name),cand) == 0)
					return;
		us_copy(&compl.candidate[compl.cnt].name,cand);
		compl.candidate[compl.cnt].isdir = isdir;
	}

	compl.first_char[cand[compl.strlen] & 0xFF] = 1;

	if (compl.cnt == 0) {
		cand0 = USTR(compl.candidate[0].name);
		compl.clen = strlen(cand0) - compl.strlen;
	}
	else
		for (c = 0; c < compl.clen ; c++)
			if (cand[compl.strlen + c] != cand0[compl.strlen + c]) {
				compl.clen = c;
				break;
			}

	compl.cnt++;
}

static void
complete_environ(void)
{
	int i;
	size_t es;
	char *pvar;
	static USTRING buff = { 0, 0 };

	for (i = 0; environ[i]; i++) {
		es = env_size[i];
		if (compl.strlen == 0 || (es >= compl.strlen &&
		  strncmp(compl.str,environ[i],compl.strlen) == 0)) {
			us_setsize(&buff,es + 1);
			pvar = USTR(buff);
			strncpy(pvar,environ[i],es);
			pvar[es] = '\0';
			register_candidate(pvar,0);
		}
	}
}

static void
complete_username(void)
{
	const char *login;

	username_find_init(compl.str,compl.strlen);
	while ( (login = username_find()) )
		/* isdir arg = 1 because we deal with user's home directories */
		register_candidate(login,1);
}

static void
complete_file(void)
{
	int isdir;
	const char *file;
	struct stat st;
	struct dirent *direntry;
	DIR *dd;

	if ( (dd = opendir(compl.dir)) == 0) {
		compl.err = errno;
		return;
	}

	pathname_set_directory(compl.dir);
	while (( direntry = readdir(dd)) ) {
		file = direntry->d_name;
		if (compl.strlen == 0) {
			if (file[0] == '.') {
				if (file[1] == '\0')
					continue;
				if (file[1] == '.' && file[2] == '\0')
					continue;
			}
		}
		else if (strncmp(file,compl.str,compl.strlen))
			continue;

		if (stat(pathname_join(file),&st) < 0)
			continue;	/* file just deleted ? */
		isdir = S_ISDIR(st.st_mode);
		if (compl.type == COMPL_TYPE_DIR && !isdir)
			continue;	/* must be a directory */
		if (compl.type == COMPL_TYPE_CMD && !isdir &&
		  (!S_ISREG(st.st_mode) ||
		  (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) == 0) )
			continue;	/* must be a directory or executable */

		register_candidate(file,isdir);
	}
	closedir(dd);
}

static void
complete_file_from_panel(void)
{
	int i, isdir;
	FILE_ENTRY *pfe;
	const char *file;

	for (i = 0; i < ppanel_file->cnt; i++) {
		pfe = ppanel_file->files[i];
		file = USTR(pfe->file);
		if (compl.strlen == 0 ?
		  pfe->flag.dotdir : strncmp(file,compl.str,compl.strlen))
			continue;

		isdir = IS_FT_DIR(pfe->file_type);
		if (compl.type == COMPL_TYPE_DIR && !isdir)
			continue;	/* must be a directory */
		if ((compl.type == COMPL_TYPE_CMD ||
		  compl.type == COMPL_TYPE_PATHCMD)	&&
		  !isdir && !IS_FT_EXEC(pfe->file_type))
			continue;	/* must be a directory or executable */

		register_candidate(file,isdir);
	}
}

static void
complete_pathcmd(void)
{
	int list, dot_in_path;
	CMD *pc;
	PATHDIR *ppd;

	command_completion_refresh();
	list = LIST_NR(*compl.str);
	dot_in_path = 0;
	for (ppd = pathdir_list_head; ppd; ppd = ppd->next)
		if (*ppd->dir == '/')
			for (pc = ppd->commands[list]; pc; pc = pc->next) {
				if (strncmp(pc->command,compl.str,compl.strlen) == 0)
					register_candidate(pc->command,0);
			}
		else 
			dot_in_path = 1;	/* must be the . directory */

	/* include subdirectories of the current directory */
	if (!dot_in_path)
		compl.type = COMPL_TYPE_DIR;	/* only DIR, not PATHCMD */
	complete_file_from_panel();
	compl.type = COMPL_TYPE_PATHCMD;
}

static int
is_cwd(void)
{
	return (compl.dirlen == 0 || strcmp(compl.dir,".") == 0 ||
	  strcmp(compl.dir,USTR(ppanel_file->dir)) == 0);
}

static void
complete_it(void)
{
	int i;

	compl.cnt = 0;
	compl.err = 0;
	for (i = 0; i < 256; i++)
		compl.first_char[i] = 0;

	if (compl.type == COMPL_TYPE_ENV)
		complete_environ();
	else if (compl.type == COMPL_TYPE_USER)
		complete_username();
	else if (compl.type == COMPL_TYPE_PATHCMD)
		complete_pathcmd();
	else if ( (compl.cwd = is_cwd()) )
		complete_file_from_panel();
	else
		complete_file();
}

static void
insert_candidate(int i)
{
	/*
	 * Note: inserting characters into the text line invalidates
	 * the compl.dir and compl.str pointers
	 */

	edit_nu_insertstr(USTR(compl.candidate[i].name) + compl.strlen,
	  textline == &line_cmd);

	if (compl.candidate[i].isdir)
		edit_nu_insertchar('/');
	else if (USTR(textline->line)[textline->curs] == ' ')
		textline->curs++;
	else
		edit_nu_insertchar(' ');

	edit_update();

	/* set file-panel cursor to the line with the candidate file */
	if (get_current_mode() == MODE_FILE_MAIN && compl.cwd &&
	  (compl.type == COMPL_TYPE_FILE || compl.type == COMPL_TYPE_DIR ||
	  compl.type == COMPL_TYPE_CMD)) {
		/* this files_find() will succeed */
		ppanel_file->curs = files_find(USTR(compl.candidate[i].name));
		pan_adjust((PANEL *)ppanel_file);
		win_panel_opt();
	}
}

const char *
compl_typestr(int uppercase)
{
	static char upper[24];
	char *src, *dst;
	int ch;

	switch (compl.type) {
	case COMPL_TYPE_FILE:
		src = "file ";
		break;
	case COMPL_TYPE_DIR:
		src = "directory ";
		break;
	case COMPL_TYPE_CMD:
	case COMPL_TYPE_PATHCMD:
		src = "command ";
		break;
	case COMPL_TYPE_USER:
		src = "user ";
		break;
	case COMPL_TYPE_ENV:
		src = "environment variable ";
		break;
	default:
		src = "";
	}

	if (!uppercase)
		return src;

	for (dst = upper; (ch = *src++ & 0xFF); )
		*dst++ = toupper(ch);
	*dst = '\0';
	return upper;
}

static void
show_results(void)
{
	char ch, *pch, *errmsg;

	if (compl.cnt == 0) {
		switch (compl.err) {
		case 0:
			errmsg = " (no match)";
			break;
		case EACCES:
			errmsg = " (permission denied)";
			break;
		case ENOENT:
			errmsg = " (no such directory)";
			break;
		default:
			errmsg = "";
		}
		win_remark("cannot complete that %sname%s",
		  compl_typestr(0),errmsg);
		return;
	}

	if (compl.cnt == 1) {
		insert_candidate(0);
		return;
	}

	if (compl.clen) {
		/* insert the common part of all candidates */
		pch = USTR(compl.candidate[0].name) + compl.strlen + compl.clen;
		ch = *pch;
		*pch = '\0';
		edit_insertstr(pch - compl.clen,1);
		*pch = ch;
		compl.strlen += compl.clen;
		/* 'compl.first_char' map is no more valid */
	}

	if (compl.cnt <= cfg_compl) {
		control_loop(MODE_COMPLETION);
		return;
	}

	win_completion(compl.cnt,compl.clen ? 0 : compl.first_char);
}

/*
 * compl_file() attempts to complete the partial text in
 * the command line
 *
 * return value:
 *    0 = completion process completed (successfully or not)
 *   <0 = nothing to complete (current word is an empty string):
 *       -1 = first word (usually the command)
 *       -2 = not the first word (usually one of the arguments)
 */
int
compl_file(void)
{
	short int first;
	char *p, *pstart, *pslash, *pend, *plinestart, *plineend;

	/*
	 * plinestart -> start of the command line
	 * plineend   -> trailing null
	 * pstart -> start of the string to be completed
	 * pend   -> immediately after the last character of the string
	 * pslash -> last slash '/' (if any)
	 */
	plineend = (plinestart = USTR(textline->line)) + textline->size;

	pstart = pend = plinestart + textline->curs;
	while (pstart > plinestart && pstart[-1] != ' ')
		pstart--;
	while (pend < plineend && *pend != ' ')
		pend++;

	if (pstart == plinestart)
		first = 1;
	else {
		first = 0;
		for (p = pstart - 1; *p == ' '; p-- )
			if (p == plinestart) {
				first = 1;
				break;
			}
		if (!first && (*p == ';' || *p == '&' ||
		  ((*p == '&' || *p == '|') && p > plinestart && p[-1] == *p)))
			first = 1;	/* after & ; && or || */
	}

	if (pstart == pend)
		/* nothing to complete */
		return first ? -1 : -2;

	for (pslash = 0, p = pend - 1; p >= pstart ; p--) {
		if (*p == '/') {
			pslash = p;
			break;
		}
	}

	/* move cursor to the end of the current word */
	textline->curs = pend - plinestart;
	edit_update_cursor();

	if (pslash == 0) {
		compl.dir = "";
		compl.dirlen = 0;
		if (*pstart == '~') {
			compl.strlen = pend - (compl.str = pstart + 1);
			compl.type = COMPL_TYPE_USER;
		}
		else if (*pstart == '$') {
			compl.strlen = pend - (compl.str = pstart + 1);
			compl.type = COMPL_TYPE_ENV;
		}
		else {
			compl.strlen = pend - (compl.str = pstart);
			compl.type = first ? COMPL_TYPE_PATHCMD : COMPL_TYPE_FILE;
		}
		complete_it();
	}
	else {
		*pslash = '\0';
		if (*pstart == '~') {
			compl.dir = dir_tilde(pstart);
			compl.dirlen = strlen(compl.dir);
		}
		else if (pstart == pslash) {
			compl.dir = "/";
			compl.dirlen = 1;
		}
		else {
			compl.dir = pstart;
			compl.dirlen = pslash - pstart;
		}
		compl.strlen = pend - (compl.str = pslash + 1);
		compl.type = first ? COMPL_TYPE_CMD : COMPL_TYPE_FILE;
		complete_it();
		*pslash = '/';
	}

	show_results();
	return 0;
}

/*
 * compl_dir() attempts to complete a directory name,
 * it is assumed that 'textline' is not empty
 */
void
compl_dir(void)
{
	char *p, *pstart, *pslash, *pend;

	/*
	 * pstart -> start of the directory name to be completed
	 * pslash -> last slash '/' (if any)
	 * pend   -> trailing null
	 */
	pend = (pstart = USTR(textline->line)) + textline->size;
	for (pslash = 0, p = pend; p >= pstart ; p--) {
		if (*p == '/') {
			pslash = p;
			break;
		}
	}

	/* move cursor to the end */
	cx_edit_end();

	if (pslash == 0) {
		compl.dir = "";
		compl.dirlen = 0;
		if (*pstart == '~') {
			compl.strlen = pend - (compl.str = pstart + 1);
			compl.type = COMPL_TYPE_USER;
		}
		else {
			compl.strlen = pend - (compl.str = pstart);
			compl.type = COMPL_TYPE_DIR;
		}
		complete_it();
	}
	else {
		*pslash = '\0';
		if (*pstart == '~') {
			compl.dir = dir_tilde(pstart);
			compl.dirlen = strlen(compl.dir);
		}
		else if (pstart == pslash) {
			compl.dir = "/";
			compl.dirlen = 1;
		}
		else {
			compl.dir = pstart;
			compl.dirlen = pslash - pstart;
		}
		compl.strlen = pend - (compl.str = pslash + 1);
		compl.type = COMPL_TYPE_DIR;
		complete_it();
		*pslash = '/';
	}

	show_results();
}

void
cx_compl_complete(void)
{
	if (panel_compl.curs >= 0)
		insert_candidate(panel_compl.curs);
	next_mode = MODE_SPECIAL_RETURN;
}
