/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* clex.h */
#include <stdlib.h>			/* qsort */
#include <string.h>			/* strcmp() */

#include "clex.h"
#include "directory.h"

#include "cfg.h"			/* config_num() */
#include "completion.h"		/* compl_dir() */
#include "control.h"		/* get_current_mode() */
#include "edit.h"			/* edit_insertstr() */
#include "filepanel.h"		/* changedir() */
#include "inout.h"			/* win_panel() */
#include "panel.h"			/* pan_adjust() */
#include "ustring.h"		/* us_copy() */
#include "userdata.h"		/* dir_tilde() */
#include "util.h"			/* emalloc() */

#define TOPNDIRS	5		/* 1 to 9, see dir_main_prepare() */

/*
 * The directory list 'dirlist' maintained here contains:
 *   a) the names of recently visited directories, it is
 *      a source for the directory panel 'panel_dir_browse'
 *   b) the last cursor position in all those directories, allowing
 *      to restore the expected cursor position when the user re-enters
 *      an already visited directory
 */
typedef struct {
	USTRING dirname;		/* directory name */
	USTRING savefile;		/* file panel's current file */
	int savetop, savecurs;	/* top line, cursor line */
} DIR_ENTRY;

static DIR_ENTRY **dirlist;	/* list of visited directories */
static int dircnt = 0;		/* number of valid entries in 'dirlist' */
static int cfgdir;			/* number of allocated entries */

void
dir_initialize(void)
{
	int i;
	DIR_ENTRY *storage;

	/*
	 * config parameters can be edited, that's why we must
	 * save the current value of the CFG_DIRECTORIES and use
	 * that value only
	 */
	cfgdir = config_num(CFG_DIRECTORIES);
	dirlist = emalloc(cfgdir * sizeof(DIR_ENTRY *));
	storage = emalloc(cfgdir * sizeof(DIR_ENTRY));
	for (i = 0; i < cfgdir; i++) {
		dirlist[i] = storage + i;
		US_INIT(dirlist[i]->dirname);
		US_INIT(dirlist[i]->savefile);
	}

	panel_dir_browse.type = PANEL_TYPE_DIR_BROWSE;
	panel_dir_browse.min = -1;
	panel_dir_browse.dir = emalloc(cfgdir * sizeof(struct dir_entry));

	panel_dir_split.type = PANEL_TYPE_DIR_SPLIT;
	panel_dir_split.min = -1;

	US_INIT(line_dir.line);
	edit_setprompt(&line_dir,"Change directory: ");
}

#define DIR_EQUAL		3
#define DIR_A_IN_B		2
#define DIR_B_IN_A		1
#define DIR_DIFFERENT	0
/*
 * function cmp_dir() compares two absolute directory names
 * and returns one of DIR_XXX:
 *   DIR_EQUAL     = 'dirA' is the same directory as 'dirB'
 *   DIR_A_IN_B    = 'dirB' is a subdirectory of 'dirA'
 *   DIR_B_IN_A    = 'dirA' is a subdirectory of 'dirB'
 *   DIR_DIFFERENT = otherwise
 *
 * Examples:
 *                   dirA        dirB          
 *   DIR_EQUAL     - /etc        /etc
 *   DIR_A_IN_B    - /var        /var/spool
 *   DIR_B_IN_A    - /usr/bin    /usr 
 *   DIR_DIFFERENT - /home/john  /home/joe
 */
static int
cmp_dir(const char *dirA, const char *dirB)
{
	int cmp, lenA, lenB, slen;

	lenA = strlen(dirA);
	lenB = strlen(dirB);

	if (lenA == lenB) {
		slen = lenA;
		cmp = DIR_EQUAL;
	}
	else if (lenA < lenB) {
		slen = lenA;
		if (dirB[slen] != '/')
			return DIR_DIFFERENT;
		cmp = DIR_A_IN_B;
	}
	else /* lenA > lenB */ {
		slen = lenB;
		if (dirA[slen] != '/')
			return DIR_DIFFERENT;
		cmp = DIR_B_IN_A;
	}
	return strncmp(dirA,dirB,slen) ? DIR_DIFFERENT : cmp;
}

/* comparison function for qsort() */
static int
qcmp(const void *e1, const void *e2)
{
	return strcmp(
	  ((struct dir_entry *)e1)->name,
	  ((struct dir_entry *)e2)->name);
}

/*
 * In order to save some CPU cycles, the panel 'panel_dir_browse'
 * is not maintained continuously. Following function builds the
 * directory panel from the 'dirlist'.
 */
void
dir_main_prepare(void)
{
	int i, j, cnt, cd;
	char *dirname;

	/*
	 * To make the directory panel more compact, directory names
     * like /a/b are considered to be "included" in /a/b/c/d and
	 * are filtered out. Top N (#define TOPNDIRS) recently
	 * visited directories form an exception to that rule, they
	 * are always present. Further their ranking (#1, #2, #3 ...)
	 * is displayed on the screen.
	 */

	for (cnt = 0, i = 0; i < dircnt; i++) {
		dirname = USTR(dirlist[i]->dirname);
		if (cnt < TOPNDIRS) {
			panel_dir_browse.dir[cnt].name = dirname;
			panel_dir_browse.dir[cnt++].rank = i + 1;
			continue;
		}

		for (j = 0; j < cnt; j++) {
			cd = cmp_dir(dirname,panel_dir_browse.dir[j].name);
			/* there are no duplicates, cannot get DIR_EQUAL */
			if (cd == DIR_A_IN_B)
				/* already present in the form of a subdirectory */
				break;
			if (cd == DIR_B_IN_A && j >= TOPNDIRS) {
				/*
				 * longer name of subdirectory replaces
				 * the shorter name we had so far
				 */
				panel_dir_browse.dir[j].name = dirname;
				break;
			}
		}
		if (j < cnt)
			continue;

		panel_dir_browse.dir[cnt].name = dirname;
		panel_dir_browse.dir[cnt++].rank = 0;
	}
	qsort(panel_dir_browse.dir,cnt,sizeof(struct dir_entry),qcmp);

	panel_dir_browse.cnt = cnt;
	panel_dir_browse.top = panel_dir_browse.curs = -1;

	/* set cursor to the entry ranked #2 (previous dir) */
	for (i = 0; i < cnt; i++)
		if (panel_dir_browse.dir[i].rank == 2) {
			panel_dir_browse.curs = i;
			pan_adjust((PANEL *)&panel_dir_browse);
			break;
		}

	panel = (PANEL *)&panel_dir_browse;
	textline = &line_dir;
	edit_nu_kill();
}

void
dir_split_prepare(void)
{
	char ch;
	int i, cnt;
	char *dirname;

	dirname = panel_dir_browse.dir[panel_dir_browse.curs].name;
	for (cnt = i = 0; (ch = dirname[i]); i++)
		if (ch == '/')
			cnt++;

	panel_dir_split.cnt = cnt + (i > 1);
	panel_dir_split.dirname = dirname;
	panel_dir_split.top = -1;
   	panel_dir_split.curs = 0;

	panel = (PANEL *)&panel_dir_split;
}

void
dir_edit_prepare(void)
{
	panel->cnt = 0;
	win_panel();
}

/* save data to 'dirlist' */
void
filepos_save(void)
{
	int new, i;
	DIR_ENTRY *x, *top;

	new = 1;
	for (top = dirlist[0], i = 0; i < cfgdir; i++) {
		x = dirlist[i];
		dirlist[i] = top;
		top = x;
		if (i == dircnt) {
			dircnt++;
			break;
		}
		if (strcmp(USTR(top->dirname),USTR(ppanel_file->dir)) == 0) {
			/* no duplicates allowed */
			new = 0;
			break;
		}
	}
	if (new)
		us_copy(&top->dirname,USTR(ppanel_file->dir));
	if (ppanel_file->cnt) {
		us_copy(&top->savefile,
		  USTR(ppanel_file->files[ppanel_file->curs]->file));
		top->savecurs = ppanel_file->curs;
		top->savetop = ppanel_file->top;
	}
	else if (new) {
		us_copy(&top->savefile,"..");
		top->savecurs = 0;
		top->savetop = 0;
	}
	dirlist[0] = top;
}

/* set the file panel cursor according to data stored in 'dirlist' */
void
filepos_set(void)
{
	char *dir;
	int i, line;
	DIR_ENTRY *pe;

	if (ppanel_file->cnt == 0)
		return;

	dir = USTR(ppanel_file->dir);
	for (i = 0; i < dircnt; i++) {
		pe = dirlist[i];
		if (strcmp(dir,USTR(pe->dirname)) == 0) {
			/* found */
			line = files_find(USTR(pe->savefile));
			ppanel_file->curs = line >= 0 ? line : pe->savecurs;
			ppanel_file->top = pe->savetop;
			pan_adjust((PANEL *)ppanel_file);
			return;
		}
	}

	/* not found */
	line = files_find("..");
	ppanel_file->curs = line >= 0 ? line : 0;
	ppanel_file->top = 0;
	pan_adjust((PANEL *)ppanel_file);
}

/* get directory length in 'panel_dir_split' line 'line' */
int
dir_split_len(int line)
{
	char ch;
	int len, level;

	level = panel_dir_split.cnt - line;
	if (level == 1)
		return 1;
	for (len = 0; (ch = panel_dir_split.dirname[len]); len++)
		if (ch == '/' && --level == 0)
			break;

	return len;
}

/* get selected directory name in MODE_DIR_SPLIT mode */
static const char *
selected_dir(void)
{
	static USTRING tmp = { 0,0 };

	if (panel_dir_split.curs == 0)
		return panel_dir_split.dirname;

	us_copyn(&tmp,panel_dir_split.dirname,
	  dir_split_len(panel_dir_split.curs));
	return USTR(tmp);
}

/* following two functions are used in all three MODE_DIR_XXX modes */
void
cx_dir_tab(void)
{
	int mode;
	const char *dir;

	mode = get_current_mode();
	if (mode == MODE_DIR_EDIT || textline->size)
		compl_dir();
	else if (panel->curs < 0)
		return;
	else {
		dir = mode == MODE_DIR_SPLIT ? selected_dir() :
		  panel_dir_browse.dir[panel_dir_browse.curs].name;
		edit_nu_insertstr(dir,0);
		if (dir[1])
			edit_nu_insertchar('/');
		edit_update();
	}
}

void
cx_dir_enter(void)
{
	int mode;
	const char *dir;

	mode = get_current_mode();
	if (textline->size)
		dir = dir_tilde(USTR(textline->line));
	else if (mode == MODE_DIR_EDIT || panel->curs < 0) {
		next_mode = MODE_SPECIAL_RETURN;
		return;
	}
	else if (mode == MODE_DIR_SPLIT)
		dir = selected_dir();
	else {
		/* DIR_MAIN -> DIR_SPLIT */
		next_mode = MODE_DIR_SPLIT;
		return;
	}

	if (changedir(dir) == 0) {
		/*
		 * In MODE_DIR_EDIT the file panel changes to display
		 * the contents of the new directory, but this change will
		 * be not recognized automatically as the panel type
		 * remains unchanged. Following statement will force the
		 * panel change. It is harmless in other MODE_DIR_XXX modes.
		 */
		panel = 0;
		next_mode = MODE_SPECIAL_RETURN;
	}
}
