/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* pid_t */
#include <errno.h>			/* errno */
#include <signal.h>			/* signal() */
#include <stdio.h>			/* puts() */
#include <string.h>			/* strcmp() */
#include <unistd.h>			/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 0xFF) == 0)
#endif
#ifndef WTERMSIG
# define WTERMSIG(stat_val) ((stat_val) & 0x7F)
#endif
#ifndef WCOREDUMP
# define WCOREDUMP(stat_val) ((stat_val) & 0x80)
#endif

#include "clex.h"
#include "exec.h"

#include "cfg.h"			/* config_str() */
#include "control.h"		/* get_current_mode() */
#include "edit.h"			/* edit_islong() */
#include "err.h"			/* err_exit() */
#include "inout.h"			/* win_heading() */
#include "history.h"		/* hist_save() */
#include "list.h"			/* list_directory() */
#include "tty.h"			/* tty_setraw() */
#include "util.h"			/* base_name() */
#include "userdata.h"		/* get_myshell() */
#include "ustring.h"		/* us_copy() */
#include "xterm_title.h"	/* xterm_title_set() */

extern int errno;

#define MAX_ARGS	8
static char *shell_argv[MAX_ARGS + 1];
static int cmd_index;

static int
parse_shellprog(const char *shellprog)
{
	char ch, *pch;
	static USTRING dup = { 0,0 };
	int i;

	us_copy(&dup,shellprog);
	pch = USTR(dup);

	/* split to arguments */
	for (i = 1, shell_argv[0] = pch; (ch = *pch); pch++ )
		if (ch == ' ') {
			*pch = '\0';
			if (pch[1] != ' ') {
				if (i >= MAX_ARGS)
					return -1;	/* too many arguments */
				shell_argv[i++] = pch + 1;
			}
		}
	shell_argv[i] = 0;

	while ( --i >= 1)
		if (strcmp(shell_argv[i],"<COMMAND>") == 0) {
			cmd_index = i;	/* real command belongs here */
			return 0;		/* finished ! */
		}

	return -2;				/* <COMMAND> omitted */
}

/* prepare to execute commands according to "SHELLPROG" */
void
exec_initialize(void)
{
	const char *shellprog;

	shellprog = config_str(CFG_SHELLPROG);

	if (*shellprog) {
		if (parse_shellprog(shellprog) == 0)
			return;		/* success */
		else
			win_warning("EXEC: Incorrect \"SHELLPROG\", "
			  "will use your login shell.");
	}

	/* no SHELLPROG => automatically set: loginshell -c <COMMAND> */
	((const char **)shell_argv)[0] = get_myshell();
	shell_argv[1] = "-c";
	cmd_index = 2;
	shell_argv[3] = 0;
}

void
exec_setprompt_cmd(void)
{
	int is_root;
	size_t len;
	const char *prompt, *chars;
	static USTRING cmd_prompt = { 0,0 };
	static char trailer[4] = " ? ";

	prompt = config_str(CFG_PROMPT);
	is_root = geteuid() == 0;	/* 0 or 1 */

	if (prompt[0] == '\0') {
		/* autoselect prompt */
		prompt = base_name(shell_argv[0]);
		len = strlen(prompt);
		if (len >= 2 && strcmp(prompt + len - 2,"sh") == 0)
			if (len >= 3 && prompt[len - 3] == 'c')
				chars = "%#";	/* looks like C-shell */
			else
				chars = "$#";	/* looks like Bourne shell */
		else
			chars = ">>";		/* some other shell */
		trailer[1] = chars[is_root];
	}
	else
		trailer[1] = '\0';

	us_cat(&cmd_prompt,is_root ? "ROOT " : "",prompt,trailer,(char *)0);
	edit_setprompt(&line_cmd,USTR(cmd_prompt));
}

#ifndef HAVE_STRERROR
static char *
strerror(int errcode)
{
	static buffer[32];

	sprintf(buffer,"errno = %d",errcode);
	return buffer;
}
#endif

static void
execute(const char *command)
{
	pid_t pid;
	int status, in;
	struct sigaction act;

	pid = fork();
	if (pid == -1)
		puts("CLEX: Can't create new process, "
		  "system resources are exhausted\n");
	else if (pid == 0) {
		act.sa_handler = SIG_DFL;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGINT,&act,0);
		sigaction(SIGQUIT,&act,0);
		((const char **)shell_argv)[cmd_index] = command;
		execv(shell_argv[0],shell_argv);
		err_exit("Can't execute shell %s (%s)",
		  shell_argv[0],strerror(errno));
	}
	else {
		while (waitpid(pid,&status,0) < 0)
			/* ignore EINTR */;
		if (WIFEXITED(status))
			if (WEXITSTATUS(status) == 0)
				printf("\nCommand successful. ");
			else
				printf("\nExit code = %d. ",WEXITSTATUS(status));
		else {
			printf("\nAbnormal termination, signal %d",
			  WTERMSIG(status));
#ifdef SYS_SIGLIST_DECLARED
			printf(" (%s)",sys_siglist[WTERMSIG(status)]);
#endif
			if (WCOREDUMP(status))
				fputs(", core image dumped",stdout);
			putchar('\n');
		}
	}

	/*
	 * list directory while the user stares at the screen
	 * reading the command output, clex appears to restart
	 * very much faster. All output routines called from
	 * list_directory() must check if curses mode is active.
	 */
	list_directory();

	ppanel_file->other->expired = 1;

	fputs("Press <enter> to continue. ",stdout);
	fflush(stdout);
	tty_setraw();
	while ((in = getchar()) != '\n' && in != '\r')
		;
	tty_reset();

	/* X-Window terminal output looks better with a line like this */
	puts("\r\n-----------------------------------------");
}

static int
user_confirm(void)
{
	int in, confirmed;

	fputs("\nExecute the command ? (y = YES) ",stdout);
	fflush(stdout);
	tty_setraw();
	in = getchar();
	tty_reset();
	confirmed = in == 'y' || in == 'Y';
	puts(confirmed ? "yes\n" : "no\n");
	fflush(stdout);
	return confirmed;
}

/*
 * return value is the warning level:
 *   0 = no warning issued
 *   1 = only mandatory warning(s)
 *   2 = configurable warning(s)
 */
static int
print_warnings(void)
{
	static USTRING cwd = { 0,0 };
	int warn;

	/* using normal output routines (no CURSES) */

	warn = 0;

	if (get_cwd_us(&cwd) < 0) {
		printf("WARNING: current working directory "
		  "is not accessible\n");
		warn = 1;	/* this warning cannot be turned off */
	}
	else if (strcmp(USTR(ppanel_file->dir),USTR(cwd))) {
		printf("WARNING: current working directory has been renamed:\n"
		  "         previous name: %s\n         new name:      %s\n",
		  USTR(ppanel_file->dir),USTR(cwd));
		us_copy(&ppanel_file->dir,USTR(cwd));
		warn = 1;
	}

	if (config_num(CFG_WARN_RM) && edit_test_rm()) {
		printf("WARNING: rm command deletes files, please confirm\n");
		warn = 2;
	}

	/* following warnings are not appropriate in history mode */
	if (get_current_mode() == MODE_FILE_MAIN) {
		if (config_num(CFG_WARN_CD) && edit_test_cd()) {
			printf("WARNING: External cd command will NOT "
			  "change directory in CLEX\n");
			warn = 2;
		}
		if (config_num(CFG_WARN_LONG) && edit_islong()) {
			printf("WARNING: This long command did not fit on the"
			  " command line\n");
			warn = 2;
		}
	}

	fflush(stdout);
	return warn;
}

void
execute_cmd(const char *cmd)
{
	static int hint = 0;
	int warn_level;

	curses_stop();
	putchar('\n');
	puts(cmd);
	putchar('\n');
	fflush(stdout);

	warn_level = print_warnings();
	if (warn_level == 0 || user_confirm()) {
		execute(cmd);
		hist_save(cmd);
		cx_edit_kill();
	}

	curses_restart();
	xterm_title_set();
	if (!hint && warn_level >= 2) {
		win_remark("warnings can be turned off in config panel");
		hint = 1;
	}
}
