/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* clex.h */
#include <sys/stat.h>	/* fstat() */
#include <fcntl.h>		/* open() */
#include <stdio.h>		/* printf() */
#include <string.h>		/* strcmp() */
#include <unistd.h>		/* read(), close() */

#include "clex.h"
#include "help.h"

#include "cfg.h"		/* config_str() */
#include "control.h"	/* get_previous_mode() */
#include "inout.h"		/* win_panel() */
#include "panel.h"		/* pan_adjust() */
#include "start.h"		/* startup_warning() */
#include "util.h"		/* emalloc() */

#ifndef SSIZE_MAX
# define SSIZE_MAX 32767
#endif

/* limits */
#define HELP_HISTORY_SIZE	16
#define HELP_PAGE_MAXCNT	50
#define HELP_FILE_MAXSIZE	50000L

static char *no_help[] = {
	"ON_LINE HELP IS NOT AVAILABLE",
	"",
	"ERROR: Helpfile is missing or cannot be read. Please check",
	"the parameter HELPFILE in the configuration panel. Ask your",
	"system administrator to reinstall the helpfile if necessary.",
	"",
	"(press ctrl-C to return back)"
};

static char **helplines = 0;	/* helpfile split into lines */
static struct {
	const char *name;			/* internal page name */
	const char *heading;		/* heading to be displayed */
	int firstline;				/* first line = 'helplines' index */
	int size;					/* total number of lines */
} helppage[HELP_PAGE_MAXCNT];
static int pagecnt;				/* number of pages in 'helppage[]' */
static int tocnum;				/* ToC page number */

/* help history - allows only to return back to previous page */
static struct {
	int pagenum, top, curs;
} history[HELP_HISTORY_SIZE];
static int head = 0, tail = 0;	/* circular buffer indices */

/* increment/decrement (+1/-1) circular buffer index X */
#define MVINDEX(X,N) \
  do X = (X + HELP_HISTORY_SIZE + N) % HELP_HISTORY_SIZE; while(0)

static int
page2num(const char *pagename)
{
	int i;

	for (i = 0; i < pagecnt; i++)
		if (strcmp(helppage[i].name,pagename) == 0)
			return i;
	return -1;		/* no such page */
}

static void
help_error(const char *helpfile, const char *msg)
{
	static const char *file = 0;

	if (file != helpfile)
		printf("HELP: Processing helpfile \"%s\"\n",file = helpfile);
	printf("HELP: %s\n",msg);
	startup_warning();
}

/* read helpfile into memory and parse it */
static int
read_helpfile(const char *helpfile)
{
	int fd, rd, linecnt, mergelink, exitloop;
	char *ptr, *helptext;
	struct stat stbuf;
	off_t filesize, left;

	/* open the helpfile */
	if ( (fd = open(helpfile,O_RDONLY)) < 0) {
		help_error(helpfile,"Cannot open helpfile");
		/* try also default help file */
		if ((fd = open(helpfile = HELP_FILE,O_RDONLY)) < 0)
			return -1;
	}

	/* check file size */
	fstat(fd,&stbuf);	/* cannot fail with valid descriptor */
	filesize = (int)stbuf.st_size;
	if (filesize == 0 || filesize >= HELP_FILE_MAXSIZE) {
		close(fd);
		help_error(helpfile,
		  filesize ? "Helpfile is too big" : "Helpfile is empty");
		return -1;
	}

	/* read into memory */
	ptr = helptext = emalloc(filesize + 1);
	left = filesize;
	while (left > 0) {
		rd = read(fd,ptr,left > SSIZE_MAX ? SSIZE_MAX : left);
		if (rd <= 0)
			break;
		left -= rd;
		ptr += rd;
	}
	close(fd);
	if (rd < 0) {
		help_error(helpfile,"File read error");
		return -1;
	}

	/* count lines */
	if (helptext[filesize - 1] != '\n')
		helptext[filesize++] = '\n';
	for (linecnt = 0, ptr = helptext; ptr < helptext + filesize; ptr++)
		if (*ptr == '\t')
			*ptr = ' ';
		else if (*ptr == '\n' || *ptr == '\0') {
			*ptr = '\0';
			linecnt++;
		}

	/* split to lines and divide to pages */
	helplines = emalloc(linecnt * sizeof(char *));
	linecnt = pagecnt = 0;
	for (exitloop = 0, ptr = helptext; ptr < helptext + filesize; ) {
		/* handle @X= parameters */
		if (ptr[0] == '@' && ptr[1] && ptr[2] == '=') {
			switch (ptr[1]) {
			case 'V': /* version stamp */
				if (strcmp(ptr + 3,VERSION))
					help_error(helpfile,"Program and helpfile versions"
					  " do not match, information\n      in the on-"
					  "line help might be inaccurate or out-of-date");
				break;
			case 'P': /* page begin */
				if (pagecnt == HELP_PAGE_MAXCNT) {
					help_error(helpfile,"Too many help pages");
					exitloop = 1;
					break;
				}
				helppage[pagecnt].name = ptr + 3;
				helppage[pagecnt].heading = ptr + 3; /* @H= overrides */
				helppage[pagecnt].firstline = linecnt;
				helppage[pagecnt].size = 0;
				mergelink = 0;
				pagecnt++;
				break;
			case 'H': /* heading */
				helppage[pagecnt - 1].heading = ptr + 3;
				break;
			case 'L': /* link */
				helplines[linecnt++] = ptr;
				/* merge this and the following line */
				mergelink = 1;
				/* do not increment page.size++ yet */
				break;
			/* default: silently ignore unknown parameter @X= */
			}
			if (exitloop)
				break;
		}
		else if (pagecnt > 0 && *ptr != '#') {
			if (!mergelink) 
				helplines[linecnt++] = ptr;
			else
				mergelink = 0;
			helppage[pagecnt - 1].size++;
		}

		/* advance to the next line */
		while(*ptr++)
			;
	}

	/* ToC must exist */
	if ( (tocnum = page2num("ToC")) < 0) {
		help_error(helpfile,"Table of contents is missing");
		return -1;
	}

	return 0;
}

void
help_initialize(void)
{
	panel_help.type = PANEL_TYPE_HELP;
	panel_help.min = 0;

	if (read_helpfile(config_str(CFG_HELPFILE)) == 0)
		return;

	/* install faked help page */
	helplines = no_help;
	helppage[tocnum = 0].name = "ToC";
	helppage[0].heading = "HELP NOT AVAILABLE";
	helppage[0].firstline = 0;
	helppage[0].size = ARRAY_SIZE(no_help);
	pagecnt = 1;

	puts("HELP: On-line help not available");
	/* this is not another startup_warning() */
	fflush(stdout);
}

static void
set_page(int pg)
{
	panel_help.top = panel_help.curs = 0;
	panel_help.cnt = helppage[pg].size;
	panel_help.pagenum = pg;
	panel_help.heading = helppage[pg].heading;
	panel_help.txt = helplines + helppage[pg].firstline;
	win_heading();
}

static void
help_goto(const char *pagename)
{
	int pg;

	if ( (pg = page2num(pagename)) < 0) {
		/* ERROR 404 :-) */
		win_warning("HELP: help-page not found.");
		return;
	}

	/* save current page parameters */
	history[head].pagenum = panel_help.pagenum;
	history[head].top = panel_help.top;
	history[head].curs = panel_help.curs;
	MVINDEX(head,1);
	if (head == tail)
		MVINDEX(tail,1);

	/* new page */
	set_page(pg);

	win_panel();
}

void
help_prepare(void)
{
	int pg;
	const char *name;

	/* CLEX help is context sensitive */
	switch (get_previous_mode()) {
	case MODE_COMPLETION:
		name = "completion";
		break;
	case MODE_CONFIG_ADMIN:
	case MODE_CONFIG_MAIN:
	case MODE_CONFIG_EDIT_NUM:
	case MODE_CONFIG_EDIT_STR:
	case MODE_CONFIG_CONFIRM:
		name = "config";
		break;
	case MODE_DIR_EDIT:
	case MODE_DIR_MAIN:
	case MODE_DIR_SPLIT:
		name = "dir";
		break;
	case MODE_FILE_MAIN:
		name = "file";
		break;
	/* no case MODE_HELP: */
	case MODE_HIST:
		name = "history";
		break;
	case MODE_LOCATE:
		name = "locate";
		break;
	case MODE_MENU:
		name = "menu";
		break;
	case MODE_SELECT:
	case MODE_DESELECT:
		name = "select";
		break;
	case MODE_SORT:
		name = "sort";
		break;
	default:
		name = "ToC";
	}

	if ( (pg = page2num(name)) < 0) {
		win_warning("HELP: help-page not found.");
		pg = tocnum;
	}
	set_page(pg);

	panel = (PANEL *)&panel_help;
	textline = 0;
	win_remark("press <F1> to get table of contents, "
	  "ctrl-C to leave help");
}

/* follow link */
void
cx_help_link(void)
{
	const char *str;

	str = panel_help.txt[panel_help.curs];
	if (str[0] == '@' && str[1] == 'L' && str[2] == '=')
		help_goto(str + 3);
}

/* display table of contents */
void
cx_help_contents(void)
{
	if (panel_help.pagenum != tocnum)
		help_goto("ToC");
}

/* go back to previous page */
void cx_help_back(void)
{
	if (head == tail) {
		win_remark("there is no previous help-page");
		return;
	}

	MVINDEX(head,-1);
	set_page(history[head].pagenum);
	panel_help.top = history[head].top;
	panel_help.curs = history[head].curs;
	pan_adjust((PANEL *)&panel_help);
	win_panel();
}
