/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* clex.h */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcpy() */

#include "clex.h"
#include "select.h"

#include "control.h"	/* get_current_mode() */
#include "edit.h"		/* edit_putstr() */
#include "inout.h"		/* win_panel() */
#include "match.h"		/* match() */
#include "list.h"		/* list_alt_directory() */
#include "sort.h"		/* sort_files() */
#include "ustring.h"	/* us_copy() */

#define FN_SELECT	1
#define FN_DESELECT	2
#define FN_TOGGLE	3

#define PATTERN_HISTORY_SIZE	12

static USTRING **history;		/* pattern history list */
static USTRING save_pattern = { 0,0 };
static int history_cnt = 0;		/* 'history' element counter */
static int pn_index = -1;		/* index for previous/next cmd */

void
select_initialize(void)
{
	/*
	 * the size of the pattern history list could be configurable,
	 * but it is not important enough to bother the user
	 */
	history = us_alloc_array(PATTERN_HISTORY_SIZE);
}

void
select_prepare(void)
{
	static int hint1 = 0, hint2 = 0;
	char *prompt;

	panel = (PANEL *)ppanel_file;

	if (get_current_mode() == MODE_SELECT)
		prompt = "SELECT files: ";
	else
		prompt = "DESELECT files: ";
	edit_setprompt(&line_tmp,prompt);
	textline = &line_tmp;
	edit_nu_putstr("*");
	edit_adjust();

	if (!hint1) {
		win_remark("shell pattern matching special characters: "
		  "? * and [...]");
		hint1 = 1;
	}
	else if (!hint2 && history_cnt > 0) {
		win_remark("TIP: Pattern history: "
		  "Ctrl-P = previous, Ctrl-N = next");
		hint2 = 1;
	}
}

static int
selectfile(FILE_ENTRY *pfe, int function)
{
	if (!pfe->flag.select) {
		/*
		 * dot and dot-dot directories cannot be selected
		 * it is not a bug, it is a protection
		 */
		if (function != FN_DESELECT && !pfe->flag.dotdir) {
			pfe->flag.select = 1;	/* one selected entry more */
			return 1;
		}
	}
	else if (function != FN_SELECT) {
		pfe->flag.select = 0;
		return -1; 				/* one less */
	}
	return 0;					/* no change */
}

void
cx_select_toggle(void)
{
	if (ppanel_file->cnt == 0)
		return;

	ppanel_file->selected +=
	  selectfile(ppanel_file->files[ppanel_file->curs],FN_TOGGLE);

	/* cursor down */
	if (ppanel_file->curs != ppanel_file->cnt - 1 &&
	  ++ppanel_file->curs - ppanel_file->top >= display.panlines)
		ppanel_file->top++;

	win_panel_opt();
}

static void
process_all(int function)
{
	int i;

	for (i = 0; i < ppanel_file->cnt; i++)
		ppanel_file->selected +=
		  selectfile(ppanel_file->files[i],function);
	win_panel();
}

void cx_select_inverse(void)	{ process_all(FN_TOGGLE);   }
void cx_select_allfiles(void)	{ process_all(FN_SELECT);   }
void cx_select_nofiles(void)	{ process_all(FN_DESELECT); }

static void
hist_save(const char *cmd)
{
	int new, i;
	USTRING *x, *top;

	new = 1;
	top = history[0];
	for (i = 0; i < PATTERN_HISTORY_SIZE ; i++) {
		x = history[i];
		history[i] = top;
		top = x;
		if (i == history_cnt) {
			history_cnt++;
			break;
		}
		if (strcmp(PUSTR(top),cmd) == 0) {
			/* avoid duplicates */
			new = 0;
			break;
		}
	}
	if (new)
		us_copy(top,cmd);
	history[0] = top;
}

void
cx_select_next(void)
{
	if (pn_index == -1) {
		win_remark("top of pattern history list");
		return;
	}

	if (pn_index-- == 0)
		edit_putstr(USTR(save_pattern));
	else
		edit_putstr(PUSTR(history[pn_index]));
}

void
cx_select_prev(void)
{
	if (pn_index >= history_cnt - 1) {
		win_remark("bottom of pattern history list");
		return;
	}

	if (++pn_index == 0)
		us_copy(&save_pattern,USTR(line_tmp.line));
	edit_putstr(PUSTR(history[pn_index]));
}

void
cx_select_files(void)
{
	int i, select;
	const char *sre;
	FILE_ENTRY *pfe;

	if (line_tmp.size == 0) {
		next_mode = MODE_SPECIAL_RETURN;
		return;
	}
	sre = USTR(textline->line);
	if (match_sre(sre) < 0)
		/* return back and correct the error */
		return;

	/* save to history list */
	pn_index = -1;
	if (strcmp(sre,"*"))
		/* don't save the asterisk, it is the default expression */
		hist_save(sre);

	select = get_current_mode() == MODE_SELECT;
	for (i = 0; i < ppanel_file->cnt; i++) {
		pfe = ppanel_file->files[i];
		if (select) {
			if(!pfe->flag.select && match(USTR(pfe->file)))
				ppanel_file->selected += selectfile(pfe,FN_SELECT);
			}
		else /* deselect */
			if (pfe->flag.select && match(USTR(pfe->file)))
				ppanel_file->selected += selectfile(pfe,FN_DESELECT);
	}
	win_panel();
	next_mode = MODE_SPECIAL_RETURN;
}

static int
qcmp(const void *e1, const void *e2)
{
	return strcmp(USTR((*(FILE_ENTRY **)e1)->file),
	  USTR((*(FILE_ENTRY **)e2)->file));
}

/* compare directories */
void
cx_select_compare(void)
{
	int min, med, max, cmp, i, cnt1;
	const char *name;
	FILE_ENTRY *pfe1, *pfe2;

	/* reread panel#2 if necessary */
	if (ppanel_file->other->expired)
		list_alt_directory();

	/* sort panel#1 by name (prepare for binary search) */
	if ( (cnt1 = ppanel_file->cnt) )
		qsort(ppanel_file->files,cnt1,sizeof(FILE_ENTRY *),qcmp);

	/* select all files in the panel#1 */
	for (i = 0; i < cnt1; i++)
		ppanel_file->selected +=
		  selectfile(ppanel_file->files[i],FN_SELECT);

	/* for all files in panel#2 search for matching file in panel#1 */
	for (i = 0; i < ppanel_file->other->cnt; i++) {
		pfe2 = ppanel_file->other->files[i];
		name = USTR(pfe2->file);
		for (pfe1 = 0, min = 0, max = cnt1 - 1; min <= max; ) {
			med = (min + max) / 2;
			cmp = strcmp(name,USTR(ppanel_file->files[med]->file));
			if (cmp == 0) {
				pfe1 = ppanel_file->files[med];
				/* entries *pfe1 and *pfe2 have the same name */
				break;
			}
			if (cmp < 0)
				max = med - 1;
			else
				min = med + 1;
		}

		if (pfe1) {
			/* files must be of the same type */
			if (pfe1->file_type != pfe2->file_type)
				pfe1 = 0;
			else if (IS_FT_DEV(pfe1->file_type)) {
				/* devices must have same major and minor numbers */
				if (strcmp(pfe1->sd_str.device,pfe2->sd_str.device))
					pfe1 = 0;
			}
			else if (IS_FT_PLAIN(pfe1->file_type)) {
				/* plain files must have the same size */
				if (pfe1->size != pfe2->size)
					pfe1 = 0;	
			}
		}

		if (pfe1) {
			/* pair of matching files */
			ppanel_file->selected +=
			  selectfile(pfe1,FN_DESELECT);
			ppanel_file->other->selected +=
			  selectfile(pfe2,FN_DESELECT);
		}
		else
			ppanel_file->other->selected += selectfile(pfe2,FN_SELECT);
	}

	/* restore the sort order */
	sort_files();

	win_remark("NOTE: contents of the files were NOT compared");
	win_panel();
}
