/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>  /* clex.h */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcmp() */

/* major() */
#ifdef MAJOR_IN_MKDEV
# include <sys/mkdev.h>
#endif
#ifdef MAJOR_IN_SYSMACROS
# include <sys/sysmacros.h>
#endif

#include "clex.h"
#include "sort.h"

#include "cfg.h"		/* config_num() */
#include "directory.h"	/* filepos_save() */
#include "panel.h"		/* pan_adjust() */
#include "ustring.h"	/* USTRING() */

void
sort_initialize(void)
{
	static char *menulines[] = {
		"sort by name",
		"sort by .extension",	/* extension or suffix ?? */
		"sort by time",
		"sort by size",
		"sort by reversed name  (use in sendmail queue directory)"
	};			/* must exactly match SORT_XXX #defines */

	panel_sort.type = PANEL_TYPE_SORT;
	panel_sort.min = -1;
	panel_sort.cnt = ARRAY_SIZE(menulines);
	panel_sort.menu = menulines;
}

void
sort_prepare(void)
{
	panel_sort.top = -1;
	panel_sort.curs = ppanel_file->sortorder;
	panel = (PANEL *)&panel_sort;
	pan_adjust(panel);
	textline = 0;
}

void
cx_sort_set(void)
{
	if (panel_sort.curs >= 0) {
		ppanel_file->sortorder = panel_sort.curs;
		filepos_save();
		sort_files();
		filepos_set();
	}
	next_mode = MODE_SPECIAL_RETURN;
}

static int group_files;
#define CMP(V1,V2) ((V1) == (V2) ? 0 : (V1) < (V2) ? -1 : 1)

/* compare reversed strings */
static int
revstrcmp(const char *s1, const char *s2)
{
	int i1, i2, c1, c2, cmp;

	for (i1 = strlen(s1),i2 = strlen(s2); i1 > 0 && i2 > 0;) {
		c1 = s1[--i1] & 0xFF;
		c2 = s2[--i2] & 0xFF;
		/*
		 * ignoring LOCALE, this sort order has nothing to do
		 * with a human language, it is intended for sendmail
		 * queue directories
		 */
		if ( (cmp = CMP(c1,c2)) )
			return cmp;
	}
	return CMP(i1,i2);
}

/* sort_group() return values (grouping order 1 < 2 < ... < 6 < 7) */
#define GROUP_DOTDIR	1	/* same as flag.dotdir ! */
#define GROUP_DOTDOTDIR	2	/* same as flag.dotdir ! */
#define GROUP_DIR		3
#define GROUP_BDEV		4
#define GROUP_CDEV		5
#define GROUP_OTHER		6
#define GROUP_PLAIN		7
static int
sort_group(FILE_ENTRY *pfe)
{
	int type;

	type = pfe->file_type;
	if (IS_FT_PLAIN(type))
		return GROUP_PLAIN;
	if (IS_FT_DIR(type)) {
		if (pfe->flag.dotdir)
			return pfe->flag.dotdir;
		return GROUP_DIR;
	}
	if (group_files > 1) {
		if (type == FT_DEV_CHAR)
			return GROUP_CDEV;
		if (type == FT_DEV_BLOCK)
			return GROUP_BDEV;
	}
	return GROUP_OTHER;
}

static int
qcmp(const void *e1, const void *e2)
{
	int cmp, group1, group2;
	const char *str1, *str2;
	time_t time1, time2;
	off_t size1, size2;

	/* I. file type grouping */
	if (group_files) {
		group1 = sort_group(*(FILE_ENTRY **)e1);
		group2 = sort_group(*(FILE_ENTRY **)e2);
		/* CMP is a macro, no function calls as arguments ! */
		if ( (cmp = CMP(group1,group2)) )
			return cmp;
	}

	/* special sorting for devices */
	if (group_files > 1 &&
	  (group1 == GROUP_BDEV || group1 == GROUP_CDEV) ) {
		/*
		 * numbers in 'sd_str.device' are properly formatted, i.e.
		 * first major, then minor, fixed position, left padded
		 */
		str1 = (*(FILE_ENTRY **)e1)->sd_str.device;
		str2 = (*(FILE_ENTRY **)e2)->sd_str.device;
		if ( (cmp = strcmp(str1,str2)) )
			return cmp;
	}

	/* II. sort order */
	switch (ppanel_file->sortorder) {
	case SORT_EMAN:
		str1 = USTR((*(FILE_ENTRY **)e1)->file);
		str2 = USTR((*(FILE_ENTRY **)e2)->file);
		return revstrcmp(str1,str2);
	case SORT_SUFFIX:
		str1 = (*(FILE_ENTRY **)e1)->extension;
		str2 = (*(FILE_ENTRY **)e2)->extension;
		cmp = STRCOLL(str1,str2);
		break;
	case SORT_TIME:
		time1 = (*(FILE_ENTRY **)e1)->mtime;
		time2 = (*(FILE_ENTRY **)e2)->mtime;
		cmp = CMP(time2,time1);
		break;
	case SORT_SIZE:
		size1 = (*(FILE_ENTRY **)e1)->size;
		size2 = (*(FILE_ENTRY **)e2)->size;
		cmp = CMP(size1,size2);
		break;
	default:	/* SORT_NAME */
		cmp = 0;
	}
	if (cmp)
		return cmp;

	/* III. sort by file name */
	str1 = USTR((*(FILE_ENTRY **)e1)->file);
	str2 = USTR((*(FILE_ENTRY **)e2)->file);
	return STRCOLL(str1,str2);
}

void
sort_files(void)
{
	if (ppanel_file->cnt == 0)
		return;

	group_files = config_num(CFG_GROUP_FILES);
	qsort(ppanel_file->files,ppanel_file->cnt,
	  sizeof(FILE_ENTRY *),qcmp);
}
