/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

/*
 * Routines in userdata.c mostly provide access to data stored
 * in /etc/passwd and /etc/group. All required data is read
 * into memory to speed up lookups. This data is valid only
 * until next data refresh, the caller must copy the returned
 * string value if it need to keep it.
 */

#include <autoconf.h>

#include <sys/stat.h>		/* stat() */
#include <sys/types.h>		/* getpwent() */
#include <sys/utsname.h>	/* uname() */
#include <time.h>			/* time() */
#include <grp.h>			/* getgrent() */
#include <pwd.h>			/* getpwent() */
#include <string.h>			/* strcmp() */
#include <stdlib.h>			/* qsort() */
#include <unistd.h>			/* stat() */

#include "clex.h"
#include "userdata.h"

#include "err.h"		/* err_exit() */
#include "ustring.h"	/* us_cat() */
#include "util.h"		/* emalloc() */

typedef struct pwdata {
	struct pwdata *next;	/* tmp pointer */
	char *login;
	char *homedir;
	uid_t uid;
} PWDATA;

typedef struct grdata {
	struct grdata *next;	/* tmp pointer */
	char *group;
	gid_t gid;
} GRDATA;

typedef struct {
	time_t timestamp;		/* when was the data obtained */
	int cnt;				/* # of entries */
	PWDATA *mydata;			/* current user */
	PWDATA **by_name;		/* sorted by name */
	PWDATA **by_uid;		/* sorted by uid */
} USERDATA;

typedef struct {
	time_t timestamp;		/* when was the data obtained */
	int cnt;				/* # of entries */
	GRDATA **by_gid;		/* sorted by gid */
} GROUPDATA;

static USERDATA  utable = { 0,0,0,0,0 };
static GROUPDATA gtable = { 0,0,0 };

static struct {
	const char *str;
	size_t len;
	int index;
} find;						/* used by username_find() */

static char *myshell = 0;

static int
qcmp_name(const void *e1, const void *e2)
{
	return strcmp((*(PWDATA **)e1)->login,(*(PWDATA **)e2)->login);
}

static int
qcmp_uid(const void *e1, const void *e2)
{
	return (*(PWDATA **)e1)->uid - (*(PWDATA **)e2)->uid;
}

static int
qcmp_gid(const void *e1, const void *e2)
{
	return (*(GRDATA **)e1)->gid - (*(GRDATA **)e2)->gid;
}

static void
read_utable(void)
{
	int i, cnt;
	uid_t my_uid;
	PWDATA *ud, *list;
	struct passwd *pw;

	if (utable.timestamp) {
		for (i = 0; i < utable.cnt; i++) {
			free(utable.by_name[i]->login);
			free(utable.by_name[i]->homedir);
			free(utable.by_name[i]);
		}
		free(utable.by_name);
		free(utable.by_uid);
	}
	utable.timestamp = time(0);

	utable.mydata = 0;
	my_uid = getuid();
	list = 0;
	for (cnt = 0; (pw = getpwent()); cnt++) {
		ud = emalloc(sizeof(PWDATA));
		ud->login = estrdup(pw->pw_name);
		ud->homedir = estrdup(pw->pw_dir);
		ud->uid = pw->pw_uid;
		if (ud->uid == my_uid && utable.mydata == 0) {
			if (myshell == 0)
				myshell = estrdup(pw->pw_shell);
			utable.mydata = ud;
		}
		ud->next = list;
		list = ud;
	}
	endpwent();

	if (cnt == 0)
		err_exit("Cannot read password file /etc/passwd");
	if (utable.mydata == 0)
		err_exit("Cannot find your entry in /etc/passwd");

	/* temporary linked list -> two sorted arrays */
	utable.by_name = emalloc(cnt * sizeof(PWDATA *));
	utable.by_uid  = emalloc(cnt * sizeof(PWDATA *));
	for (ud = list, i = 0; i < cnt; i++, ud = ud->next)
		utable.by_name[i] = utable.by_uid[i] = ud;
	qsort(utable.by_name,cnt,sizeof(PWDATA *),qcmp_name);
	qsort(utable.by_uid ,cnt,sizeof(PWDATA *),qcmp_uid );
	utable.cnt = cnt;
}

static void
read_gtable(void)
{
	int i, cnt;
	GRDATA *gd, *list;
	struct group *gr;

	if (gtable.timestamp) {
		for (i = 0; i < gtable.cnt; i++) {
			free(gtable.by_gid[i]->group);
			free(gtable.by_gid[i]);
		}
		free(gtable.by_gid);
	}
	gtable.timestamp = time(0);

	list = 0;
	for (cnt = 0; (gr = getgrent()); cnt++) {
		gd = emalloc(sizeof(GRDATA));
		gd->group = estrdup(gr->gr_name);
		gd->gid = gr->gr_gid;
		gd->next = list;
		list = gd;
	}
	endgrent();

	if (cnt == 0)
		err_exit("Cannot read user group file /etc/group");

	/* temporary linked list -> sorted array */
	gtable.by_gid  = emalloc(cnt * sizeof(GRDATA *));
	for (gd = list, i = 0; i < cnt; i++, gd = gd->next)
		gtable.by_gid[i] = gd;
	qsort(gtable.by_gid,cnt,sizeof(GRDATA *),qcmp_gid);
	gtable.cnt = cnt;
}

/*
 * This initialization function can and should be called
 * repeatedly to refresh the cached data.
 */
int
userdata_initialize(void)
{
	int reloaded;
	struct stat st;

	reloaded = 0;
	if (stat("/etc/passwd",&st) < 0 ||
	  st.st_mtime >= utable.timestamp) {
		read_utable();
		reloaded = 1;
	}
	if (stat("/etc/group" ,&st) < 0 ||
	  st.st_mtime >= gtable.timestamp) {
		read_gtable();
		reloaded = 1;
	}

	return reloaded;
}

const char *
get_myhomedir(void)
{
	return utable.mydata->homedir;
}

const char *
get_mylogin(void)
{
	return utable.mydata->login;
}

/* an exception: get_myshell() always returns the same string */
const char *
get_myshell(void)
{
	return myshell;
}

const char *
get_homedir(const char *user)
{
	int min, med, max, cmp;

	for (min = 0, max = utable.cnt - 1; min <= max; ) {
		med = (min + max) / 2;
		cmp = strcmp(user,utable.by_name[med]->login);
		if (cmp == 0)
			return utable.by_name[med]->homedir;
		if (cmp < 0)
			max = med - 1;
		else
			min = med + 1;
	}
	return 0;
}

/*
 * this customized version of get_homedir() accepts empty string
 * argument and requires string length 'len' - as needed by dir_tilde()
 */
static const char *
get_homedir2(const char *user, size_t len)
{
	int min, med, max, cmp;

	if (len == 0)
		return get_myhomedir();

	for (min = 0, max = utable.cnt - 1; min <= max; ) {
		med = (min + max) / 2;
		cmp = strncmp(user,utable.by_name[med]->login,len);
		if (cmp == 0)
			return utable.by_name[med]->homedir;
		if (cmp < 0)
			max = med - 1;
		else
			min = med + 1;
	}
	return 0;
}

const char *
get_login(uid_t uid)
{
	int min, med, max, cmp;

	for (min = 0, max = utable.cnt - 1; min <= max; ) {
		med = (min + max) / 2;
		cmp = uid - utable.by_uid[med]->uid;
		if (cmp == 0)
			return utable.by_uid[med]->login;
		if (cmp < 0)
			max = med - 1;
		else
			min = med + 1;
	}
	return 0;
}

const char *
get_group(gid_t gid)
{
	int min, med, max, cmp;

	for (min = 0, max = gtable.cnt - 1; min <= max; ) {
		med = (min + max) / 2;
		cmp = gid - gtable.by_gid[med]->gid;
		if (cmp == 0)
			return gtable.by_gid[med]->group;
		if (cmp < 0)
			max = med - 1;
		else
			min = med + 1;
	}
	return 0;
}

/*
 * Following two functions implement username completion. First
 * username_find_init() is called to initialize the search, thereafter
 * each call to username_find() returns one matching entry in
 * alphabetical order.
 */
void
username_find_init(const char *str, size_t len)
{
	int min, med, max, cmp;

	find.str = str;
	find.len = len;

	if (len == 0) {
		find.index = 0;
		return;
	}

	for (min = 0, max = utable.cnt - 1; min <= max; ) {
		med = (min + max) / 2;
		cmp = strncmp(str,utable.by_name[med]->login,len);
		if (cmp == 0) {
			/*
			 * the binary search algorithm is slightly altered here,
			 * there may be more matches than one and we need to
			 * find the first one.
			 */
			if (min == max) {
				find.index = med;
				return;
			}
			max = med;
		}
		else if (cmp < 0)
			max = med - 1;
		else
			min = med + 1;
	}

	find.index = utable.cnt;
}

const char *
username_find(void)
{
	const char *login;

	if (find.index >= utable.cnt)
		return 0;
	login = utable.by_name[find.index]->login;
	if (find.len && strncmp(find.str,login,find.len))
		return 0;
	find.index++;
	return login;
}

const char *
get_machinename(void)
{
#ifndef HAVE_UNAME
	return "localhost";
#else
	int i;
	static const char *machinename = 0;
	struct utsname ut;

	if (machinename == 0) {
		uname(&ut);
		for (i = 0; ut.nodename[i]; i++)
			if (ut.nodename[i] == '.') {
				ut.nodename[i] = '\0';
				break;
			}
		machinename = estrdup(ut.nodename);
	}
	return machinename;
#endif
}

/* get user@host string */
const char *
get_user_at_host(void)
{
	static USTRING uh = { 0,0 };

	us_cat(&uh,get_mylogin(),"@",get_machinename(),(char *)0);
	return USTR(uh);
}

/*
 * dir_tilde() function performs tilde substitution. It understands
 * ~user/dir notation and transforms it to proper directory name.
 * The result of the substitution (if performed) is stored in
 * a static buffer that might get overwritten by successive calls.
 */
const char
*dir_tilde(const char *dir)
{
	char ch;
	int i;
	const char *home;
	static USTRING buff = { 0,0 };

	if (dir[0] != '~')
		return dir;

	for (i = 1; (ch = dir[i]) && ch != '/'; i++)
		;
	home = get_homedir2(dir + 1,i - 1);
	if (home == 0)
		return dir;		/* no such user */

	us_cat(&buff,home,dir + i,(char *)0);
	return USTR(buff);
}
