/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>			/* clex.h */
#include <errno.h>				/* errno */
#include <stdarg.h>				/* va_list */
#include <stdlib.h>				/* free() */
#include <string.h>				/* strlen() */
#include <unistd.h>				/* getcwd() */

#include "clex.h"
#include "ustring.h"

#include "util.h"				/* emalloc() */

extern int errno;

/*
 * USTRING structure (defined in clex.h) can hold a string
 * of not limited length. The memory is allocated when needed
 * (saving memory) and not deallocated between uses (saving CPU time)
 *
 * Functions defined below are used to manipulate USTRINGs.
 * Each USTRING must be first initialized either using US_INIT()
 * macro or by assigning zero to the 'memalloc' structure element.
 *
 * Note: The value of the 'str' element is not defined before
 * the USTRING structure is used in any of the us_xxx() functions.
 * Afterwards it is guaranteed to be a valid non-null pointer.
 */

/* these are tunable parameters */
#define ALLOC_UNIT			24	/* in bytes */
#define MINIMUM_FREE		(6 * ALLOC_UNIT)

/*
 * SHOULD_CHANGE_ALLOC() is true if
 *  1) we need more memory, or
 *  2) we can free considerable amount of memory (MINIMUM_FREE)
 */
#define SHOULD_CHANGE_ALLOC(RQ)	\
	(pustr->memalloc < RQ || pustr->memalloc >= RQ + MINIMUM_FREE)

/* memory is allocated in chunks to prevent excessive resizing */
#define ROUND_ALLOC(RQ)	\
	((1 + (RQ - 1) / ALLOC_UNIT) * ALLOC_UNIT)

/* us_setsize() makes room for 'memreq' characters */
void
us_setsize(USTRING *pustr, size_t memreq)
{
	if (SHOULD_CHANGE_ALLOC(memreq)) {
		if (pustr->memalloc)
			free(pustr->str);
		pustr->memalloc = ROUND_ALLOC(memreq);
		pustr->str = emalloc(pustr->memalloc);
	}
}

/* like us_resize(), but preserving contents */
void
us_resize(USTRING *pustr, size_t memreq)
{
	if (pustr->memalloc == 0)
		us_setsize(pustr,memreq);
	else if (SHOULD_CHANGE_ALLOC(memreq)) {
		pustr->memalloc = ROUND_ALLOC(memreq);
		pustr->str = erealloc(pustr->str,pustr->memalloc);
	}
}

void
us_copy(USTRING *pustr, const char *src)
{
	us_setsize(pustr,strlen(src) + 1);
	strcpy(pustr->str,src);
}

/* note: us_copyn() adds terminating null */
void
us_copyn(USTRING *pustr, const char *src, size_t len)
{
	us_setsize(pustr,len + 1);
	strncpy(pustr->str,src,len);
	pustr->str[len] = '\0';
}

/* usage: us_cat(&ustring, str1, str2, ..., strN, (char *)0); */
void
us_cat(USTRING *pustr, ...)
{
	size_t len;
	char *str;
	va_list argptr;

	va_start(argptr, pustr);
	for (len = 1; (str = va_arg(argptr, char *)); )
		len += strlen(str);
	va_end(argptr);

	us_setsize(pustr,len);

	va_start(argptr, pustr);
	for (len = 0; (str = va_arg(argptr, char *)); ) {
		strcpy(pustr->str + len,str);
		len += strlen(str);
	}
	va_end(argptr);
}

USTRING **
us_alloc_array(int n)
{
	int i;
	USTRING **ppustr, *pustr;

	if (n <= 0)
		return 0;

	ppustr = emalloc(n * sizeof(USTRING *));
	pustr = emalloc(n * sizeof(USTRING));
	for (i = 0; i < n; i++)
		(ppustr[i] = pustr + i)->memalloc = 0;

	return ppustr;
}

/* USTRING version of getcwd() */
int
get_cwd_us(USTRING *pustr)
{
	if (pustr->memalloc == 0)
		 us_setsize(pustr,2 * ALLOC_UNIT);

	for (;/* until return*/;) {
		if (getcwd(pustr->str,pustr->memalloc))
			return 0;
		if (errno == ERANGE)
			/* any other reasonable value instead of 3 is also OK */
			us_setsize(pustr,pustr->memalloc + 3 * ALLOC_UNIT);
		else
			return -1;
	}
}

/* USTRING version of readlink() */
int
get_link_us(USTRING *pustr, const char *path)
{
#ifndef HAVE_READLINK
	return -1;
#else
	int chars;

	if (pustr->memalloc == 0)
		 us_setsize(pustr,2 * ALLOC_UNIT);

	for (;/* until return*/;) {
		chars = readlink(path,pustr->str,pustr->memalloc);
		if (chars < 0)
			return -1;
		if (chars == pustr->memalloc)
			/* very probably truncated */
			us_setsize(pustr,pustr->memalloc + 3 * ALLOC_UNIT);
		else {
			pustr->str[chars] = '\0';
			return 0;
		}
	}
#endif
}
