/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* clex.h */
#include <signal.h>		/* signal() */
#include <stdio.h>		/* printf() */
#include <stdlib.h>		/* getenv() */
#include <string.h>		/* strchr() */
#include <unistd.h>		/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "clex.h"
#include "xterm_title.h"

#include "cfg.h"		/* config_num() */
#include "userdata.h"	/* get_mylogin() */
#include "util.h"		/* estrdup() */

static int enabled = 0;
static const char *oldtitle;

#define XPROP_TIMEOUT	10

/*
 * run the command
 *   xprop -id $WINDOWID WM_NAME
 * to get the current xterm title
 *
 * errors are not reported, we don't want to confuse the user
 */
static const char
*get_title(void)
{
	int rd, fd[2];
	char *wid, *buff, *p1, *p2;
	char title[128];
	size_t left;
	pid_t pid;

	if ( (wid = getenv("WINDOWID")) == 0)
		return 0;

	if (pipe(fd) < 0 || (pid = fork()) < 0)
		return 0;

	if (pid == 0) {
		/* child */
		close(fd[0]);	/* close read end */
		if (fd[1] != STDOUT_FILENO) {
			if (dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
				_exit(126);
			close(fd[1]);
		}
		signal(SIGALRM,SIG_DFL);
		alarm(XPROP_TIMEOUT);
		execlp("xprop", "xprop", "-id", wid, "WM_NAME", (char *)0);
		_exit(127);
	}

	/* parent continues here */
	/* do not return before waitpid() or you'll create a zombie */
	close(fd[1]);
	buff = title;
	left = sizeof(title) - 1 /* 1 byte for trailing null */;
	while (left > 0) {
		rd = read(fd[0],buff,left);
		if (rd <= 0)
			break;
		left -= rd;
		buff += rd;
	}
	*buff = '\0';
	close(fd[0]);

	if (waitpid(pid, 0, 0) < 0)
		return 0;

	/* rd < 0 means an error, rd > 0 means buffer too small */
	if (rd)
		return 0;

	/* get the window title in quotation marks */
	p1 = strchr(title,'\"');
	if (p1 == 0)
		return 0;
	p2 = strchr(++p1,'\"');
	if (p2 == 0)
		return 0;
	*p2 = '\0';

	return estrdup(p1);
}

/* xterm_title_initialize() may be called repeatedly */
void
xterm_title_initialize(void)
{
	int xterm_feature;
	char *term;
	static int got_title = 0;

	enabled = 0;
	xterm_feature = config_num(CFG_XTERM_TITLE);
	/* 0 = disabled */
	if (xterm_feature == 0)
		return;
	/* 1 = automatic -> check $TERM for xterm, kterm or dtterm */
	if (xterm_feature == 1) {
		if ( (term = getenv("TERM")) == 0)
			return;
		if (strncmp(term,"xterm",5) && strncmp(term,"kterm",5) &&
		  strncmp(term,"dtterm",6))
			return;
	}
	/* otherwise 2 = enabled */
	enabled = 1;
	if (got_title == 0) {
		if ( (oldtitle = get_title()) == 0)
			oldtitle = "terminal";
		got_title = 1;
	}
}

void
xterm_title_set(void)
{
	/* Xterm FAQ claims \007 is incorrect, but everybody uses it */
	if (enabled)
		printf("\033]0;clex: %s\007",get_user_at_host());
}

void
xterm_title_restore(void)
{
	if (enabled)
		printf("\033]0;%s\007",oldtitle);
}
